// Robot finite state machine example.
// Fill in the body of function updateRobot() below to guide robot to each open square.
// 
// See also: http://nifty.stanford.edu/2010/dodds-picobot/

// Variables defining the map size:
final int SIZE = 20; // Width of each square in pixels
final int ROWS = 11; // Number of squares in height
final int COLS = 21; // Number of squares in width

int FRAMERATE = 30; // Frames drawn per second

int MAPTYPE = 0; // Try 0, 1, 2, or 3.

// Names for each possible robot state (add new entries here if necessary):
final int START = 0;

// Function to turn robot right.
void turnRight() {
  robot.turnRight();
}

// Function to turn robot left.
void turnLeft() {
  robot.turnLeft();
}

// This function defines state transitions and can tell the robot to change direction.
// Parameters: 
// f, r, b, l: each is true if the forward, right, back, or left square is occupied.
// state: current robot state.
// Returns: new robot state.
int updateRobot(boolean f, boolean r, boolean b, boolean l, int state) {
    // Fill in with instructions on how to turn and change state
	// Use calls to turnLeft() and turnRight() to move robot.
    return state; 
}

// Sorry, below is messy and uncommented... check back later.

void keyPressed() {
    switch(key) {
    case ' ': 
        MAPTYPE = (MAPTYPE + 1) % 4;
        initAll();
        break;
    case '1':
        FRAMERATE = max(1, FRAMERATE - 10);
        frameRate(FRAMERATE);
        break;
    case '2':
        FRAMERATE = FRAMERATE + 10;
        frameRate(FRAMERATE);
        break;
    }
}

final int GOOD = 0, BAD = 1, DONE = 2;
int board[][];
Robot robot;

void setup() {
  size(COLS * SIZE, ROWS * SIZE);
  frameRate(FRAMERATE);
  initAll();
}

void initAll() {
    initBoard();
    initRobot();
}

void initBoard() {
  board = new int[ROWS][COLS];
  // Fill in border cells:
  for (int i = 0; i < COLS; ++i) 
    board[0][i] = board[ROWS - 1][i] = BAD;
  for (int i = 0; i < ROWS; ++i)
    board[i][0] = board[i][COLS - 1] = BAD;
    
  if (MAPTYPE == 1) {
    int start = 2;
    for (int i = 0; i < COLS; i += 2) {
      for (int r = 0; r < ROWS - 3; ++r)
        board[r + start][i] = BAD;
      start = 3 - start;
    }
  }
  
  if (MAPTYPE == 2) {
    for (int r = 0; r < ROWS; ++r) 
      for (int c = 0; c < COLS; ++c)
        board[r][c] = BAD;
    randomBoard((int)random(1, ROWS-1), (int)random(1, COLS-1));
  }
  
  if (MAPTYPE == 3) {
    int r0 = (int)random(2, ROWS - 2);
    int r1 = (int)random(2, ROWS - 2);
    int c0 = (int)random(2, COLS - 2);
    int c1 = (int)random(2, COLS - 2);
    for (int r = min(r0, r1); r <= max(r0, r1); ++r)
      for (int c = min(c0, c1); c <= max(c0, c1); ++c)
        board[r][c] = BAD;
  }
}

void initRobot() {
  // Choose random initial position for robot
  int r0, c0;
  do {
    r0 = (int)random(ROWS);
    c0 = (int)random(COLS);
  } while (board[r0][c0] == BAD);
  robot = new Robot(r0, c0, Robot.NORTH);
}

int neighbors(int r, int c) {
  int n = 0;
  if (board[r+1][c] != BAD) n++;
  if (board[r-1][c] != BAD) n++;
  if (board[r][c+1] != BAD) n++;
  if (board[r][c-1] != BAD) n++;
  return n;
}

void randomBoard(int r, int c) {
  board[r][c] = GOOD;
  int dirs[] = {0,1,2,3};
  for (int i = 0; i < 4; ++i) {
    int tmp = dirs[i];
    int j = (int)random(i,4);
    dirs[i] = dirs[j];
    dirs[j] = tmp;
  }
  for (int d : dirs) {
    int rr = 0, cc = 0;
    if (d == 0) { rr = 0; cc = 1; }
    if (d == 1) { rr = 0; cc = -1; }
    if (d == 2) { rr = 1; cc = 0; }
    if (d == 3) { rr = -1; cc = 0; }
    int r1 = r + rr;
    int c1 = c + cc;
    if (r1 > 0 && r1 < ROWS - 1 && c1 > 0 && c1 < COLS - 1 && neighbors(r1, c1) == 1)
      randomBoard(r+rr, c+cc);
  }
}

void drawBoard() {
  for (int r = 0; r < ROWS; ++r) {
    for (int c = 0; c < COLS; ++c) {
      int b = board[r][c];
      switch(b) {
        case GOOD:
          fill(255);
          break;
        case BAD:
          fill(0);
          break;
        case DONE:
          fill(125);
          break;
      }
      rect(c * SIZE, r * SIZE, SIZE, SIZE);
    }
  }
}

class Robot {
  static final int NORTH = 0, EAST = 1, SOUTH = 2, WEST = 3;
  int r;
  int c;
  int state;
  int dir; 
  Robot(int r0, int c0, int dir0) {
    state = START;
    dir = dir0 % 4;
    r = r0;
    c = c0;
  }
  void turnLeft() {
    dir = (dir + 3) % 4;
  }
  void turnRight() {
    dir = (dir + 1) % 4;
  }
  void draw() {
    fill(0, 0, 200);
    rect(c * SIZE, r * SIZE, SIZE, SIZE);
    fill(200, 0, 0);
    int w = SIZE/4;
    int xoff = 0, yoff = 0;
    if (dir == NORTH)
      yoff = -SIZE / 2;
    else if (dir == EAST)
      xoff = SIZE / 2;
    else if (dir == SOUTH)
      yoff = SIZE / 2;
    else if (dir == WEST)
      xoff = -SIZE / 2;
    rect(c * SIZE + SIZE/2 - w/2 + xoff, r * SIZE + SIZE/2 - w/2 + yoff, w, w);
  }
  void update() {
    boolean blocked[] = {false, false, false, false};
    blocked[NORTH] = board[r-1][c] == BAD;
    blocked[EAST] = board[r][c+1] == BAD;
    blocked[SOUTH] = board[r+1][c] == BAD;
    blocked[WEST] = board[r][c-1] == BAD;
    println(blocked);
    println(dir);
    blocked = arot(blocked, dir);
    println(blocked);
    state = updateRobot(blocked[0], blocked[1], blocked[2], blocked[3], state);
  }
  
  boolean[] arot(boolean a[], int r) {
    int n = a.length;
    boolean ar[] = new boolean[n];
    for (int i = 0; i < n; ++i) {
      ar[i] = a[(i+r)%n];
    }
    return ar;
  }

  void move() {
    board[r][c] = DONE;
    int rnew = r;
    int cnew = c;
    if (dir == NORTH)
      rnew = r - 1;
    else if (dir == SOUTH)
      rnew = r + 1;
    else if (dir == EAST)
      cnew = c + 1;
    else if (dir == WEST)
      cnew = c - 1;
    if (board[rnew][cnew] != BAD) {
      r = rnew;
      c = cnew;
    }
  }
}


void draw() {
  robot.update();
  robot.move();
  
  drawBoard();
  robot.draw();
}
